#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
"""
The simplest possible panel server. To launch a panel server containing the intake gui
run:

panel serve intake/interface/server.py

"""

import intake.interface
intake.interface.instance.servable()



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import os
from intake.utils import make_path_posix


def unzip(f, outpath):
    import zipfile
    z = zipfile.ZipFile(f, 'r')
    z.extractall(outpath)
    out = [make_path_posix(os.path.join(outpath, fn.filename))
           for fn in z.filelist]
    z.close()
    return out


def untargz(f, outpath):
    import tarfile
    tar = tarfile.open(f, "r:gz")
    out = [make_path_posix(os.path.join(outpath, fn.name))
           for fn in tar.getmembers()]
    tar.extractall(outpath)
    tar.close()
    return out


def untarbz(f, outpath):
    import tarfile
    tar = tarfile.open(f, "r:bz2")
    out = [make_path_posix(os.path.join(outpath, fn.name))
           for fn in tar.getmembers()]
    tar.extractall(outpath)
    tar.close()
    return out


def untar(f, outpath):
    import tarfile
    tar = tarfile.open(f, "r:")
    out = [make_path_posix(os.path.join(outpath, fn.name))
           for fn in tar.getmembers()]
    tar.extractall(outpath)
    tar.close()
    return out


def ungzip(f, outpath):
    import gzip
    z = gzip.open(f)
    fn = os.path.basename(f)[:-3]
    with open(os.path.join(outpath, fn), 'wb') as fout:
        data = True
        while data:
            data = z.read(2**15)
            fout.write(data)
    return [make_path_posix(os.path.join(outpath, fn))]


def unbzip(f, outpath):
    import bz2
    z = bz2.open(f)
    fn = os.path.basename(f)[:-3]
    with open(os.path.join(outpath, fn), 'wb') as fout:
        data = True
        while data:
            data = z.read(2 ** 15)
            fout.write(data)
    return [make_path_posix(os.path.join(outpath, fn))]


decomp = {'zip': unzip,
          'tgz': untargz,
          'tbz': untarbz,
          'tar': untar,
          'gz': ungzip,
          'bz': unbzip}



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from . import base, import_name


class TextFilesSource(base.DataSource):
    """Read textfiles as sequence of lines

    Prototype of sources reading sequential data.

    Takes a set of files, and returns an iterator over the text in each of them.
    The files can be local or remote. Extra parameters for encoding, etc.,
    go into ``storage_options``.
    """
    name = 'textfiles'
    version = '0.0.1'
    container = 'python'
    partition_access = True

    def __init__(self, urlpath, text_mode=True, text_encoding='utf8',
                 compression=None, decoder=None, read=True, metadata=None,
                 storage_options=None):
        """

        Parameters
        ----------
        urlpath : str or list(str)
            Target files. Can be a glob-path (with "*") and include protocol
            specified (e.g., "s3://"). Can also be a list of absolute paths.
        text_mode : bool
            Whether to open the file in text mode, recoding binary
            characters on the fly
        text_encoding : str
            If text_mode is True, apply this encoding. UTF* is by far the most
            common
        compression : str or None
            If given, decompress the file with the given codec on load. Can
            be something like "gzip", "bz2", or to try to guess from the filename,
            'infer'
        decoder : function, str or None
            Use this to decode the contents of files. If None, you will get
            a list of lines of text/bytes. If a function, it must operate on
            an open file-like object or a bytes/str instance, and return a
            list
        read : bool
            If decoder is not None, this flag controls whether bytes/str get
            passed to the function indicated (True) or the open file-like
            object (False)
        storage_options: dict
            Options to pass to the file reader backend, including text-specific
            encoding arguments, and parameters specific to the remote
            file-system driver, if using.
        """
        self._urlpath = urlpath
        self._storage_options = storage_options or {}
        self._dataframe = None
        self._files = None
        if isinstance(decoder, str):
            decoder = import_name(decoder)
        self.decoder = decoder
        self.compression = compression
        self.mode = 'rt' if text_mode else 'rb'
        self.encoding = text_encoding
        self._read = read

        super(TextFilesSource, self).__init__(metadata=metadata)

    def _get_schema(self):
        from fsspec import open_files
        if self._files is None:

            urlpath = self._get_cache(self._urlpath)[0]

            self._files = open_files(
                urlpath, mode=self.mode, encoding=self.encoding,
                compression=self.compression,
                **self._storage_options)
            self.npartitions = len(self._files)
        return base.Schema(dtype=None,
                           shape=(None, ),
                           npartitions=self.npartitions,
                           extra_metadata=self.metadata)

    def _get_partition(self, i):
        return get_file(self._files[i], self.decoder, self._read)

    def read(self):
        self._get_schema()
        return self.to_dask().compute()

    def to_spark(self):
        from intake_spark.base import SparkHolder
        h = SparkHolder(False, [
            ('textFile', (self._urlpath, ))
        ], {})
        return h.setup()

    def to_dask(self):
        import dask.bag as db
        from dask import delayed
        self._get_schema()
        dfile = delayed(get_file)
        return db.from_delayed([dfile(f, self.decoder, self._read)
                                for f in self._files])


def get_file(f, decoder, read):
    """Serializable function to take an OpenFile object and read lines"""
    with f as f:
        if decoder is None:
            return list(f)
        else:
            d = f.read() if read else f
            out = decoder(d)
            if isinstance(out, (tuple, list)):
                return out
            else:
                return [out]



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
from distutils.version import LooseVersion

try:
    import dfviz
    assert LooseVersion(dfviz.__version__) >= LooseVersion("0.1.0")
except ImportError:
    dfviz = False

try:
    import xrviz
    from xrviz.dashboard import Dashboard as XRViz
    assert LooseVersion(xrviz.__version__) >= LooseVersion("0.1.1")
except ImportError:
    xrviz = False

import panel as pn
from ..base import BaseView
from ...catalog.local import LocalCatalogEntry


class Plots(BaseView):
    """
    Panel for displaying pre-defined plots from catalog.

    Parameters
    ----------
    source: intake catalog entry, or list of same
        source to describe in this object

    Attributes
    ----------
    has_plots: bool
        whether the source has plots defined
    instructions_contents: str
        instructions to put on the plot selector label
    options: list
        plots options defined on the source
    selected: str
        name of selected plot
    children: list of panel objects
        children that will be used to populate the panel when visible
    panel: panel layout object
        instance of a panel layout (row or column) that contains children
        when visible
    watchers: list of param watchers
        watchers that are set on children - cleaned up when visible
        is set to false.
    """
    select = None

    def __init__(self, source=None, **kwargs):
        self.custom = pn.widgets.Button(name='Customize...', width_policy='min', align='center')
        self.source = source
        self.panel = pn.Column(name='Plot', width_policy='max', margin=0)
        super().__init__(**kwargs)

    def setup(self):
        self.instructions = pn.pane.Markdown(
            self.instructions_contents, align='center', width_policy='max')
        self.select = pn.widgets.Select(options=self.options, height=30,
                                        align='center', min_width=200)
        self.pane = pn.pane.HoloViews(self._plot_object(self.selected),
                                      name="Plot")

        self.watchers = [
            self.select.param.watch(self.callback, ['options', 'value']),
            self.custom.param.watch(self.interact, ['clicks'])
        ]
        self.out = pn.Row(self.pane, name="Plot")

        self.children = [
            pn.Row(
                self.instructions,
                self.select,
                self.custom
            ),
            self.out,
        ]

    @BaseView.source.setter
    def source(self, source):
        """When the source gets updated, update the the options in
        the selector"""
        if source and isinstance(source, list):
            source = source[0]
        if isinstance(source, LocalCatalogEntry):
            source = source()
        BaseView.source.fset(self, source)
        if self.select:
            self.select.options = self.options
        if source and dfviz and source.container == 'dataframe':
            self.custom.disabled = False
        elif source and xrviz and source.container in ['xarray', 'ndarray', 'numpy']:
            self.custom.disabled = False
        else:
            self.custom.disabled = True

    @property
    def has_plots(self):
        """Whether the source has plots defined"""
        return self.source is not None and len(self._source.plots) > 0

    @property
    def instructions_contents(self):
        """Instructions to put on the plot selector label"""
        if self.has_plots:
            return '**Select from the predefined plots:**'
        return '*No predefined plots found - declare these in the catalog*'

    @property
    def options(self):
        """Plots options defined on the source"""
        return (['None'] + self.source.plots) if self.source is not None else []

    @property
    def selected(self):
        """Name of selected plot"""
        return self.select.value if self.select is not None else None

    @selected.setter
    def selected(self, selected):
        """When plot is selected set, make sure widget stays upto date"""
        self.select.value = selected

    def callback(self, *events):
        for event in events:
            if event.name == 'value':
                self.pane.object = self._plot_object(event.new)
            if event.name == 'options':
                self.instructions.object = self.instructions_contents

    def interact(self, _):
        # "customize" was pressed
        if self.selected == 'None':
            kwargs = {}
        else:
            kwargs = self.source.metadata['plots'][self.selected]
        if self.source.container == 'dataframe':
            df = self.source.to_dask()
            if df.npartitions == 1:
                df = df.compute()
            viz = dfviz.DFViz(df, **kwargs)
        elif self.source.container in ['xarray', 'ndarray', 'numpy']:
            import xarray
            try:
                data = self.source.to_dask()
            except NotImplementedError:
                data = self.source.read()
            if not isinstance(data, (xarray.DataArray, xarray.Dataset)):
                data = xarray.DataArray(data)
            viz = XRViz(data, **kwargs)
        else:
            return
        self.out[0] = viz.panel

    def _plot_object(self, selected):
        if selected and str(selected) != "None":
            plot_method = getattr(self.source.plot, selected)
            self.out[0] = self.pane
            if plot_method:
                return plot_method()

    def __getstate__(self, include_source=True):
        """Serialize the current state of the object. Set include_source
        to False when using with another panel that will include source."""
        state = super().__getstate__(include_source)
        state.update({
            'selected': self.selected,
        })
        return state

    def __setstate__(self, state):
        """Set the current state of the object from the serialized version.
        Works inplace. See ``__getstate__`` to get serialized version and
        ``from_state`` to create a new object."""
        super().__setstate__(state)
        if self.visible:
            self.selected = state.get('selected')
        return self



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
from distutils.version import LooseVersion

gl = globals()


def do_import():
    error = too_old = False
    try:
        import panel as pn
        too_old = LooseVersion(pn.__version__) < LooseVersion("0.9.5")
    except ImportError as e:
        error = e

    if too_old or error:
        raise RuntimeError("Please install panel to use the GUI `conda "
                           "install -c conda-forge panel>=0.8.0`. Import "
                           "failed with error: %s" % error)

    from .gui import GUI
    css = """
    .scrolling {
      overflow: scroll;
    }
    """
    pn.config.raw_css.append(css)  # add scrolling class from css (panel GH#383, GH#384)
    pn.extension()
    gl['instance'] = GUI()


def __getattr__(attr):
    if attr == 'instance':
        do_import()
    return gl['instance']


def output_notebook(inline=True, logo=False):
    """
    Load the notebook extension

    Parameters
    ----------
    inline : boolean (optional)
        Whether to inline JS code or load it from a CDN
    logo : boolean (optional)
        Whether to show the logo(s)
    """
    try:
        import hvplot
    except ImportError:
        raise ImportError("The intake plotting API requires hvplot."
                          "hvplot may be installed with:\n\n"
                          "`conda install -c pyviz hvplot` or "
                          "`pip install hvplot`.")
    import holoviews as hv
    return hv.extension('bokeh', inline=inline, logo=logo)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from collections.abc import MutableMapping
import pkgutil
import warnings
import importlib
import inspect
import time
import logging

import entrypoints

from ..config import conf
logger = logging.getLogger('intake')


class DriverSouces:
    """
    Handles the various ways in which drivers can be known to Intake
    """

    def __init__(self, config=None, do_scan=None):
        """

        Parameters
        ----------
        config: intake.config.Config oinstance, optional
            If not given, will use ``intake.config.conf`` singleton
        do_scan: bool or None
            Whether to scan packages with names ``intake_*`` for valid drivers.
            This is for backward compatibility only. If not given, value comes
            from the config key "package_scan", default False.
        """
        self.conf = config or conf
        self.do_scan = do_scan
        self._scanned = None
        self._entrypoints = None
        self.registered = {}

    @property
    def package_scan(self):
        return self.conf.get("package_scan", False) if self.do_scan is None else self.do_scan

    @package_scan.setter
    def package_scan(self, val):
        self.conf["package_scan"] = val

    def from_entrypoints(self):
        if self._entrypoints is None:
            # must cache this, since lookup if fairly slow and we do this a lot
            self._entrypoints = list(entrypoints.get_group_all("intake.drivers"))
        return self._entrypoints

    def from_conf(self):
        return [
            entrypoints.EntryPoint(
                name=k,
                module_name=v.rsplit(":", 1)[0] if ":" in v else v.rsplit(".", 1)[0],
                object_name=v.rsplit(":", 1)[1] if ":" in v else v.rsplit(".", 1)[1]
            )
            for k, v in self.conf.get("drivers", {}).items() if v
        ]

    def disabled(self):
        return [k for k, v in self.conf.get("drivers", {}).items() if v is False]

    def __setitem__(self, key, value):
        super(DriverSouces, self).__setitem__(key, value)
        self.save()

    def __delitem__(self, key):
        super(DriverSouces, self).__delitem__(key)
        self.save()

    @property
    def scanned(self):
        # cache since imports should not change during session
        if self._scanned is None:
            # these are already classes
            self._scanned = _package_scan()
        return self._scanned

    def enabled_plugins(self):
        # priority order (decreasing): runtime, config, entrypoints, package scan
        out = {}
        if self.package_scan:
            out.update(self.scanned)

        for ep in self.from_entrypoints() + self.from_conf():
            out[ep.name] = ep

        out.update(self.registered)
        out = {k: v for k, v in out.items() if k not in self.disabled()}
        return out

    def register_driver(self, name, value, clobber=False, do_enable=False):
        """Add runtime driver definition

        Parameters
        ----------
        name: str
            Name of the driver
        value: str, entrypoint or class
            Pointer to the implementation
        clobber: bool
            If True, perform the operation even if the driver exists
        do_enable: bool
            If True, unset the disabled flag for this driver
        """
        if not clobber and name in self.enabled_plugins():
            raise ValueError(f"Driver {name} already enabled")
        if name in self.disabled():
            if do_enable:
                self.enable(name)
            else:
                logger.warning(f"Adding driver {name}, but it is disabled")

        self.registered[name] = value

    def unregister_driver(self, name):
        """Remove runtime registered driver"""
        self.registered.pop(name)

    def enable(self, name, driver=None):
        """
        Explicitly assign a driver to a name, or remove ban

        Updates the associated config, which will be persisted

        Parameters
        ----------
        name : string
            As in ``'zarr'``
        driver : string
            Dotted object name, as in ``'intake_xarray.xzarr.ZarrSource'``.
            If None, simply remove driver disable flag, if it is found
        """
        config = self.conf
        if 'drivers' not in config:
            config['drivers'] = {}
        if driver:
            config['drivers'][name] = driver
        elif config['drivers'].get(name) is False:
            del config['drivers'][name]
        config.save()

    def disable(self, name):
        """Disable a driver by name.

        Updates the associated config, which will be persisted

        Parameters
        ----------
        name : string
            As in ``'zarr'``
        """
        config = self.conf
        if 'drivers' not in config:
            config['drivers'] = {}
        config['drivers'][name] = False
        config.save()


def _load_entrypoint(entrypoint):
    """
    Call entrypoint.load() and, if it fails, raise context-specific errors.
    """
    try:
        return entrypoint.load()
    except ImportError as err:
        raise ConfigurationError(
            f"Failed to load {entrypoint.name} driver because module "
            f"{entrypoint.module_name} could not be imported.") from err
    except AttributeError as err:
        raise ConfigurationError(
            f"Failed to load {entrypoint.name} driver because no object "
            f"named {entrypoint.object_name} could be found in the module "
            f"{entrypoint.module_name}.") from err


def _package_scan(path=None, plugin_prefix='intake_'):
    """Scan for intake drivers and return a dict of plugins.

    This searches path (or sys.path) for packages with names that start with
    plugin_prefix.  Those modules will be imported and scanned for subclasses
    of intake.source.base.Plugin.  Any subclasses found will be instantiated
    and returned in a dictionary, with the plugin's name attribute as the key.
    """
    warnings.warn("Package scanning may be removed", category=PendingDeprecationWarning)

    plugins = {}

    for importer, name, ispkg in pkgutil.iter_modules(path=path):
        if name.startswith(plugin_prefix):
            t = time.time()
            new_plugins = load_plugins_from_module(name)

            for plugin_name, plugin in new_plugins.items():
                if plugin_name in plugins:
                    orig_path = inspect.getfile(plugins[plugin_name])
                    new_path = inspect.getfile(plugin)
                    warnings.warn('Plugin name collision for "%s" from'
                                  '\n    %s'
                                  '\nand'
                                  '\n    %s'
                                  '\nKeeping plugin from first location.'
                                  % (plugin_name, orig_path, new_path))
                else:
                    plugins[plugin_name] = plugin
            logger.debug("Import %s took: %7.2f s" % (name, time.time() - t))
    return plugins


drivers = DriverSouces()


def load_plugins_from_module(module_name):
    """Imports a module and returns dictionary of discovered Intake plugins.

    Plugin classes are instantiated and added to the dictionary, keyed by the
    name attribute of the plugin object.
    """
    from intake.source import DataSource
    from intake.catalog import Catalog
    plugins = {}

    try:
        try:
            mod = importlib.import_module(module_name)
        except ImportError as error:
            if module_name.endswith('.py'):
                # Provide a specific error regarding the removal of behavior
                # that intake formerly supported.
                raise ImportError(
                    "Intake formerly supported executing arbitrary Python "
                    "files not on the sys.path. This is no longer supported. "
                    "Drivers must be specific with a module path like "
                    "'package_name.module_name, not a Python filename like "
                    "'module.py'.") from error
            else:
                raise
    except Exception as e:
        logger.debug("Import module <{}> failed: {}".format(module_name, e))
        return {}
    for _, cls in inspect.getmembers(mod, inspect.isclass):
        # Don't try to register plugins imported into this module elsewhere
        if issubclass(cls, (Catalog, DataSource)):
            plugins[cls.name] = cls

    return plugins


class ConfigurationError(Exception):
    pass





#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import appdirs
import json
import os
import subprocess
import sys

from intake.config import conf
from intake.utils import make_path_posix
from .local import YAMLFilesCatalog, Catalog


def load_user_catalog():
    """Return a catalog for the platform-specific user Intake directory"""
    cat_dir = user_data_dir()
    if not os.path.isdir(cat_dir):
        return Catalog()
    else:
        return YAMLFilesCatalog(cat_dir)


def user_data_dir():
    """Return the user Intake catalog directory"""
    return appdirs.user_data_dir(appname='intake', appauthor='intake')


def load_global_catalog():
    """Return a catalog for the environment-specific Intake directory"""
    cat_dir = global_data_dir()
    if not os.path.isdir(cat_dir):
        return Catalog()
    else:
        return YAMLFilesCatalog(cat_dir)


CONDA_VAR = 'CONDA_PREFIX'
VIRTUALENV_VAR = 'VIRTUAL_ENV'


def conda_prefix():
    """Fallback: ask conda in PATH for its prefix"""
    try:
        out = subprocess.check_output(['conda', 'info', '--json'])
        return json.loads(out.decode())["default_prefix"]
    except (subprocess.CalledProcessError, json.JSONDecodeError, OSError):
        return False


def which(program):
    """Emulate posix ``which``"""
    import distutils.spawn
    return distutils.spawn.find_executable(program)


def global_data_dir():
    """Return the global Intake catalog dir for the current environment"""
    prefix = False
    if VIRTUALENV_VAR in os.environ:
        prefix = os.environ[VIRTUALENV_VAR]
    elif CONDA_VAR in os.environ:
        prefix = sys.prefix
    elif which('conda'):
        # conda exists but is not activated
        prefix = conda_prefix()

    if prefix:
        # conda and virtualenv use Linux-style directory pattern
        return make_path_posix(os.path.join(prefix, 'share', 'intake'))
    else:
        return appdirs.site_data_dir(appname='intake', appauthor='intake')


def load_combo_catalog():
    """Load a union of the user and global catalogs for convenience"""
    user_dir = user_data_dir()
    global_dir = global_data_dir()
    desc = 'Generated from data packages found on your intake search path'
    cat_dirs = []
    if os.path.isdir(user_dir):
        cat_dirs.append(user_dir + '/*.yaml')
        cat_dirs.append(user_dir + '/*.yml')
    if os.path.isdir(global_dir):
        cat_dirs.append(global_dir + '/*.yaml')
        cat_dirs.append(global_dir + '/*.yml')
    for path_dir in conf.get('catalog_path', []):
        if path_dir != '':
            if not path_dir.endswith(('yaml', 'yml')):
                cat_dirs.append(path_dir + '/*.yaml')
                cat_dirs.append(path_dir + '/*.yml')
            else:
                cat_dirs.append(path_dir)

    return YAMLFilesCatalog(cat_dirs, name='builtin', description=desc)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
'''

'''
import logging
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from os.path import dirname, exists, join
import os, shutil

# External imports

# Intake imports
from intake.cli.util import Subcommand

#-----------------------------------------------------------------------------
# API
#-----------------------------------------------------------------------------

class Example(Subcommand):
    ''' Create example catalog

    '''

    name = "example"

    def initialize(self):
        pass

    def invoke(self, args):
        print('Creating example catalog...')
        files = ['us_states.yml', 'states_1.csv', 'states_2.csv']
        for filename in files:
            if exists(filename):
                print('Cannot create example catalog in current directory.\n'
                    '%s already exists.' % filename)
                return 1

        src_dir = join(dirname(__file__), '..', '..', 'sample')

        for filename in files:
            src_name = join(src_dir, filename)
            dest_name = filename
            dest_dir = dirname(filename)
            print('  Writing %s' % filename)
            if dest_dir != '' and not exists(dest_dir):
                os.mkdir(dest_dir)
            shutil.copyfile(src_name, dest_name)

        print('''\nTo load the catalog:
    >>> import intake
    >>> cat = intake.open_catalog('%s')
    ''' % files[0])



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from os.path import expanduser
import logging
import os
import posixpath
import yaml
from intake.utils import yaml_load

from .utils import make_path_posix
logger = logging.getLogger('intake')

confdir = make_path_posix(
    os.getenv('INTAKE_CONF_DIR', os.path.join(expanduser('~'), '.intake')))


defaults = {
    'auth': {'cls': 'intake.auth.base.BaseAuth'},
    'port': 5000,
    'cache_dir': posixpath.join(confdir, 'cache'),
    'cache_disabled': False,
    'cache_download_progress': True,
    'logging': 'INFO',
    'catalog_path': [],
    'persist_path': posixpath.join(confdir, 'persisted'),
    'package_scan': False
}


def cfile():
    return make_path_posix(
        os.getenv('INTAKE_CONF_FILE', posixpath.join(confdir, 'conf.yaml')))


class Config(dict):
    def __init__(self, filename=None, **kwargs):
        self.filename = filename if filename is not None else cfile()
        self.reload_all()
        super().__init__(**kwargs)

    def reset(self):
        """Set conf values back to defaults"""
        self.clear()
        self.update(defaults)

    def save(self):
        """Save current configuration to file as YAML

        Uses ``.filename`` for target location
        """
        if self.filename is False:
            return
        try:
            os.makedirs(os.path.dirname(self.filename))
        except (OSError, IOError):
            pass
        with open(self.filename, 'w') as f:
            yaml.dump(dict(self), f)

    def reload_all(self):
        self.reset()
        self.load()
        self.load_env()

    def load(self, fn=None):
        """Update global config from YAML file

        If fn is None, looks in global config directory, which is either defined
        by the INTAKE_CONF_DIR env-var or is ~/.intake/ .
        """
        fn = fn or self.filename

        if os.path.isfile(fn):
            with open(fn) as f:
                try:
                    self.update(yaml_load(f))
                except Exception as e:
                    logger.warning('Failure to load config file "{fn}": {e}'
                                   ''.format(fn=fn, e=e))

    def load_env(self):
        """Analyse environment variables and update conf accordingly"""
        # environment variables take precedence over conf file
        for key, envvar in [['cache_dir', 'INTAKE_CACHE_DIR'],
                            ['catalog_path', 'INTAKE_PATH'],
                            ['persist_path', 'INTAKE_PERSIST_PATH']]:
            if envvar in os.environ:
                self[key] = make_path_posix(os.environ[envvar])
        self['catalog_path'] = intake_path_dirs(self['catalog_path'])
        for key, envvar in [['cache_disabled', 'INTAKE_DISABLE_CACHING'],
                            ['cache_download_progress', 'INTAKE_CACHE_PROGRESS']]:
            if envvar in os.environ:
                self[key] = os.environ[envvar].lower() in ['true', 't', 'y', 'yes']
        if 'INTAKE_LOG_LEVEL' in os.environ:
            self['logging'] = os.environ['INTAKE_LOG_LEVEL']


def intake_path_dirs(path):
    """Return a list of directories from the intake path.

    If a string, perhaps taken from an environment variable, then the
    list of paths will be split on the character ":" for posix of ";" for
    windows. Protocol indicators ("protocol://") will be ignored.
    """
    if isinstance(path, (list, tuple)):
        return path
    import re
    pattern = re.compile(";" if os.name == 'nt' else r"(?<!:):(?![:/])")
    return pattern.split(path)


conf = Config()
conf.reload_all()
save_conf = conf.save
load_cond = conf.load

logger.setLevel(conf['logging'].upper())
ch = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - '
                              '%(filename)s:%(funcName)s:L%(lineno)d - '
                              '%(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)
logger.debug('Intake logger set to debug')



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import collections
from collections import OrderedDict
import collections.abc
import datetime
from contextlib import contextmanager
import warnings
import yaml
import sys
import os


def make_path_posix(path):
    """ Make path generic """
    if '://' in path:
        return path
    return path.replace('\\', '/').replace('//', '/')


def no_duplicates_constructor(loader, node, deep=False):
    """Check for duplicate keys while loading YAML

    https://gist.github.com/pypt/94d747fe5180851196eb
    """

    mapping = {}
    for key_node, value_node in node.value:
        key = loader.construct_object(key_node, deep=deep)
        value = loader.construct_object(value_node, deep=deep)
        if key in mapping:
            from intake.catalog.exceptions import DuplicateKeyError

            raise DuplicateKeyError("while constructing a mapping",
                                    node.start_mark,
                                    "found duplicate key (%s)" % key,
                                    key_node.start_mark)
        mapping[key] = value

    return loader.construct_mapping(node, deep)


def tuple_constructor(loader, node, deep=False):
    return tuple(loader.construct_object(node, deep=deep)
                 for node in node.value)


def represent_dictionary_order(self, dict_data):
    return self.represent_mapping('tag:yaml.org,2002:map', dict_data.items())


yaml.add_representer(OrderedDict, represent_dictionary_order)


@contextmanager
def no_duplicate_yaml():
    yaml.SafeLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        no_duplicates_constructor)
    yaml.SafeLoader.add_constructor('tag:yaml.org,2002:python/tuple',
                                    tuple_constructor)
    try:
        yield
    finally:
        yaml.SafeLoader.add_constructor(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            yaml.constructor.SafeConstructor.construct_yaml_map
        )


def yaml_load(stream):
    """Parse YAML in a context where duplicate keys raise exception"""
    with no_duplicate_yaml():
        return yaml.safe_load(stream)


def classname(ob):
    """Get the object's class's name as package.module.Class"""
    import inspect
    if inspect.isclass(ob):
        return '.'.join([ob.__module__, ob.__name__])
    else:
        return '.'.join([ob.__class__.__module__, ob.__class__.__name__])


class DictSerialiseMixin(object):

    __tok_cache = None

    def __new__(cls, *args, **kwargs):
        """Capture creation args when instantiating"""
        o = object.__new__(cls)
        o._captured_init_args = args
        o._captured_init_kwargs = kwargs

        return o

    @property
    def classname(self):
        return classname(self)

    def __dask_tokenize__(self):
        if self.__tok_cache is None:
            from dask.base import tokenize
            self.__tok_cache = tokenize(self.__getstate__())
        return self.__tok_cache

    def __getstate__(self):
        args = [arg.__getstate__() if isinstance(arg, DictSerialiseMixin)
                else arg
                for arg in self._captured_init_args]
        # We employ OrderedDict in several places. The motivation
        # is to speed up dask tokenization. When dask tokenizes a plain dict,
        # it sorts the keys, and it turns out that this sort operation
        # dominates the call time, even for very small dicts. Using an
        # OrderedDict steers dask toward a different and faster tokenization.
        kwargs = collections.OrderedDict({
            k: arg.__getstate__()
            if isinstance(arg, DictSerialiseMixin) else arg
            for k, arg in self._captured_init_kwargs.items()
        })
        return collections.OrderedDict(cls=self.classname,
                                       args=args,
                                       kwargs=kwargs)

    def __setstate__(self, state):
        # reconstitute instances here
        self._captured_init_kwargs = state['kwargs']
        self._captured_init_args = state['args']
        state.pop('cls', None)
        self.__init__(*state['args'], **state['kwargs'])

    def __hash__(self):
        from dask.base import tokenize
        return int(tokenize(self), 16)

    def __eq__(self, other):
        return hash(self) == hash(other)


def remake_instance(data):
    import importlib
    if isinstance(data, str):
        data = {'cls': data}
    else:
        data = data.copy()
    mod, klass = data.pop('cls').rsplit('.', 1)
    module = importlib.import_module(mod)
    cl = getattr(module, klass)
    return cl(*data.get('args', ()), **data.get('kwargs', {}))


def pretty_describe(object, nestedness=0, indent=2):
    """Maintain dict ordering - but make string version prettier"""
    if not isinstance(object, dict):
        return str(object)
    sep = f'\n{" " * nestedness * indent}'
    out = sep.join((f'{k}: {pretty_describe(v, nestedness + 1)}' for k, v in object.items()))
    if nestedness > 0 and out:
        return f'{sep}{out}'
    return out


def decode_datetime(obj):
    import numpy
    if not isinstance(obj, numpy.ndarray) and "__datetime__" in obj:
        try:
            obj = datetime.datetime.strptime(
                    obj["as_str"],
                    "%Y%m%dT%H:%M:%S.%f%z",
            )
        except ValueError: # Perhaps lacking tz info
            obj = datetime.datetime.strptime(
                    obj["as_str"],
                    "%Y%m%dT%H:%M:%S.%f",
            )
    return obj


def encode_datetime(obj):
    if isinstance(obj, datetime.datetime):
        return {"__datetime__": True, "as_str": obj.strftime("%Y%m%dT%H:%M:%S.%f%z")}
    return obj


class RegistryView(collections.abc.Mapping):
    """
    Wrap registry dict in a read-only dict view.

    Subclasses define attributes filled into warning and error messages:
    - self._registry_name
    - self._register_func_name
    - self._unregister_func_name
    """
    def __init__(self, registry):
        self._registry = registry

    def __repr__(self):
        return f"{self.__class__.__name__}({self._registry!r})"

    def __getitem__(self, key):
        return self._registry[key]

    def __iter__(self):
        yield from self._registry

    def __len__(self):
        return len(self._registry)

    # Support the common mutation methods for now, but warn.

    def update(self, *args, **kwargs):
        warnings.warn(
            f"In a future release of intake, the {self._registry_name} will "
            f"not be directly mutable. Use {self._register_func_name}.",
            DeprecationWarning)
        self._registry.update(*args, **kwargs)
        # raise TypeError(
        #     f"The registry cannot be edited directly. "
        #     f"Instead, use the {self._register_func_name{")

    def __setitem__(self, key, value):
        warnings.warn(
            f"In a future release of intake, the {self._registry_name} will "
            f"not be directly mutable. Use {self._register_func_name}.",
            DeprecationWarning)
        self._registry[key] = value
        # raise TypeError(
        #     f"The registry cannot be edited directly. "
        #     f"Instead, use the {self._register_func_name{")

    def __delitem__(self, key):
        warnings.warn(
            f"In a future release of intake, the {self._registry_name} will "
            f"not be directly mutable. Use {self._unregister_func_name}.",
            DeprecationWarning)
        del self._registry[key]
        # raise TypeError(
        #     f"The registry cannot be edited directly. "
        #     f"Instead, use the {self._unregister_func_name{")


class DriverRegistryView(RegistryView):
    # This attributes are used by the base class
    # to fill in warning and error messages.
    _registry_name = "intake.registry"
    _register_func_name = "intake.register_driver"
    _unregister_func_name = "intake.unregister_driver"


class ContainerRegistryView(RegistryView):
    # This attributes are used by the base class
    # to fill in warning and error messages.
    _registry_name = "intake.container_map"
    _register_func_name = "intake.register_container"
    _unregister_func_name = "intake.unregister_container"


class ModuleImporter:
    def __init__(self, destination):
        self.destination = destination
        self.module = None

    def __getattribute__(self, item):
        d = object.__getattribute__(self, "__dict__")
        if item in d:
            return d[item]
        if self.module is None:
            print("Importing module: ", self.destination)
            self.module = __import__(self.destination)
        else:
            print("Referencing module: ", self.destination)
        sys.modules[self.destination] = self.module
        return getattr(self.module, item)


def is_notebook() -> bool:
    """ Check if code is running in a notebook

    Copied from tqdm.autonotebook

    Returns
    -------
    bool
        True if inside a notebook. False if not inside a notebook.
    """

    try:
        get_ipython = sys.modules['IPython'].get_ipython
        if 'IPKernelApp' not in get_ipython().config:  # pragma: no cover
            raise ImportError("console")
        if 'VSCODE_PID' in os.environ:  # pragma: no cover
            raise ImportError("vscode")
        return True
    except Exception:
        return False



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import ast
import os
from functools import partial

import intake
import panel as pn

import fsspec
from fsspec.registry import known_implementations
from ..base import Base, MAX_WIDTH, enable_widget, ICONS


class FileSelector(Base):
    """
    Panel interface for picking files

    The current path is stored in .path and the current selection is stored in
    .url.

    Parameters
    ----------
    filters: list of string
        extentions that are included in the list of files - correspond to
        catalog extensions.
    done_callback: func, opt
        called when the object's main job has completed. In this case,
        selecting a file.

    Attributes
    ----------
    url: str
        path to local catalog file
    children: list of panel objects
        children that will be used to populate the panel when visible
    panel: panel layout object
        instance of a panel layout (row or column) that contains children
        when visible
    watchers: list of param watchers
        watchers that are set on children - cleaned up when visible
        is set to false.
    """
    def __init__(self, filters=['yaml', 'yml'], done_callback=None,  **kwargs):
        self.filters = filters
        self.panel = pn.Column(name='Local', width_policy='max', margin=0)
        self.done_callback = done_callback
        self.fs = fsspec.filesystem("file")
        super().__init__(**kwargs)

    def setup(self):
        self.path_text = pn.widgets.TextInput(value=os.getcwd() + '/',
                                              width_policy='max')
        self.protocol = pn.widgets.Select(options=list(sorted(known_implementations)),
                                          value='file', name='protocol')
        self.storage_options = pn.widgets.TextInput(name='kwargs',
                                                    value="{}")
        self.go = pn.widgets.Button(name='⇨')
        self.validator = pn.pane.SVG(None, width=25)
        self.main = pn.widgets.MultiSelect(size=15, width_policy='max')
        self.home = pn.widgets.Button(name='🏠', width=40, height=30)
        self.up = pn.widgets.Button(name='‹', width=30, height=30)

        self.make_options()

        self.watchers = [
            self.go.param.watch(self.go_clicked, 'clicks'),
            self.protocol.param.watch(self.protocol_changed, 'value'),
            #self.path_text.param.watch(self.validate, ['value']),
            #self.path_text.param.watch(self.make_options, ['value']),
            self.home.param.watch(self.go_home, 'clicks'),
            self.up.param.watch(self.move_up, 'clicks'),
            self.main.param.watch(self.move_down, ['value'])
        ]

        self.children = [
            pn.Row(self.protocol, self.storage_options),
            pn.Row(self.home, self.up, self.path_text, self.go, margin=0),
            self.main
        ]

    def protocol_changed(self, *_):
        self.path_text.value = ""
        self.main.options = []
        self.main.value = []

    def go_clicked(self, *_):
        self.fs = fsspec.filesystem(
            self.protocol.value, **ast.literal_eval(self.storage_options.value))
        self.make_options()

    @property
    def path(self):
        return self.path_text.value

    @property
    def url(self):
        """Path to local catalog file"""
        return (self.protocol.value + "://" +
                os.path.join(self.path, self.main.value[0]))

    def move_up(self, arg=None):
        self.path_text.value = self.fs._parent(self.path_text.value)
        self.make_options()

    def go_home(self, arg=None):
        self.protocol.value = 'file'
        self.path_text.value = os.getcwd() + os.path.sep

    def validate(self, arg=None):
        """Check that inputted path is valid - set validator accordingly"""
        if os.path.isdir(self.path):
            self.validator.object = None
        else:
            self.validator.object = ICONS['error']

    def make_options(self, arg=None):
        if self.done_callback:
            self.done_callback(False)
        out = []
        try:
            for f in self.fs.ls(self.path, True):
                bn = os.path.basename(f['name'].rstrip('/'))
                if bn.startswith('.'):
                    continue
                elif f['type'] == 'directory':
                    out.append(bn + '/')
                elif not self.filters or any(bn.endswith(ext) for ext in self.filters):
                    out.append(bn)
        except Exception as e:
            print(e)

        self.main.value = []
        self.main.options = sorted(out)

    def move_down(self, *events):
        for event in events:
            if event.name == 'value' and len(event.new) > 0:
                fn = event.new[0]
                if fn.endswith('/'):
                    if self.path_text.value:
                        self.path_text.value = os.path.join(self.path_text.value, fn)
                    else:
                        self.path_text.value = fn
                    self.make_options()
                elif self.done_callback:
                    self.done_callback(True)

    def __getstate__(self):
        """Serialize the current state of the object."""
        return {
            'path': self.path,
            'selected': self.main.value
        }

    def __setstate__(self, state):
        """Set the current state of the object from the serialized version.
        Works inplace. See ``__getstate__`` to get serialized version and
        ``from_state`` to create a new object."""
        self.path_text.value = state['path']
        self.main.value = state['selected']
        return self


class URLSelector(Base):
    """
    Panel interface for inputting a URL to a remote catalog

    The inputted URL is stored in .url.

    Attributes
    ----------
    url: str
        url to remote files (including protocol)
    children: list of panel objects
        children that will be used to populate the panel when visible
    panel: panel layout object
        instance of a panel layout (row or column) that contains children
        when visible
    watchers: list of param watchers
        watchers that are set on children - cleaned up when visible
        is set to false.
    """
    def __init__(self, **kwargs):
        self.panel = pn.Row(name='URL',
                            width_policy='max', margin=0)
        super().__init__(**kwargs)

    def setup(self):
        self.main = pn.widgets.TextInput(
            placeholder="Full URL with protocol",
            width_policy='max')
        self.children = ['URL:', self.main]

    @property
    def url(self):
        """URL to remote files (including protocol)"""
        return self.main.value

    def __getstate__(self):
        """Serialize the current state of the object."""
        return {'url': self.url}

    def __setstate__(self, state):
        """Set the current state of the object from the serialized version.
        Works inplace. See ``__getstate__`` to get serialized version and
        ``from_state`` to create a new object."""
        self.main.value = state['url']
        return self


class CatAdder(Base):
    """Panel for adding new cats from local file or remote

    Parameters
    ----------
    done_callback: function with cat as input
        function that is called when the "Add Catalog" button is clicked.

    Attributes
    ----------
    cat_url: str
        url to remote files or path to local files. Depends on active tab
    cat: catalog
        catalog object initialized from from cat_url
    children: list of panel objects
        children that will be used to populate the panel when visible
    panel: panel layout object
        instance of a panel layout (row or column) that contains children
        when visible
    watchers: list of param watchers
        watchers that are set on children - cleaned up when visible
        is set to false.
    """
    tabs = None

    def __init__(self, done_callback=None, **kwargs):
        self.done_callback = done_callback
        self.panel = pn.Column(name='Add Catalog',
                               width_policy='max',
                               max_width=MAX_WIDTH,
                               margin=0)
        self.widget = pn.widgets.Button(name='Add Catalog',
                                        disabled=True,
                                        width_policy='min')
        self.fs = FileSelector(done_callback=partial(enable_widget, self.widget))
        self.url = URLSelector()
        super().__init__(**kwargs)

    def setup(self):
        self.selectors = [self.fs, self.url]
        self.tabs = pn.Tabs(*map(lambda x: x.panel, self.selectors))
        self.validator = pn.pane.SVG(None, width=25)

        self.watchers = [
            self.widget.param.watch(self.add_cat, 'clicks'),
            self.tabs.param.watch(self.tab_change, 'active'),
            self.fs.main.param.watch(self.remove_error, 'value'),
            self.url.main.param.watch(self.remove_error, 'value'),
        ]

        self.children = [self.tabs, pn.Row(self.widget, self.validator)]

    @property
    def cat_url(self):
        """URL to remote files or path to local files. Depends on active tab."""
        url = self.selectors[self.tabs.active].url
        if self.selectors[self.tabs.active] is self.fs:
            fs = self.fs.fs
        else:
            fs = None
        return url, fs

    @property
    def cat(self):
        """Catalog object initialized from from cat_url"""
        # might want to do some validation in here
        url, fs = self.cat_url
        if fs:
            return intake.open_catalog(url, fs=fs)
        else:
            return intake.open_catalog(url)

    def add_cat(self, arg=None):
        """Add cat and close panel"""
        try:
            self.done_callback(self.cat)
            self.visible = False
        except Exception as e:
            self.validator.object = ICONS['error']
            raise e

    def remove_error(self, *args):
        """Remove error from the widget"""
        self.validator.object = None

    def tab_change(self, event):
        """When tab changes remove error, and enable widget if on url tab"""
        self.remove_error()
        if event.new == 1:
            self.widget.disabled = False

    def __getstate__(self):
        """Serialize the current state of the object"""
        return {
            'visible': self.visible,
            'local': self.fs.__getstate__(),
            'remote': self.url.__getstate__(),
            'active': self.tabs.active if self.tabs else 0
        }

    def __setstate__(self, state):
        """Set the current state of the object from the serialized version.
        Works inplace. See ``__getstate__`` to get serialized version and
        ``from_state`` to create a new object."""
        self.fs.__setstate__(state['local'])
        self.url.__setstate__(state['remote'])
        self.visible = state.get('visible', True)
        if self.visible:
            self.tabs.active = state['active']
        return self
